"""The models for the package `mock-event-generator`."""
from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class Upload:
    """Class containing the information required to submit a file to GraceDB."""

    message: str
    tags: set[str]
    filename: str
    delay: float


@dataclass
class GEventDescription:
    """Class containing all the information required to create a G-event."""

    id: str
    source: str
    pipeline: str
    group: str
    search: str
    offline: bool
    gpstime: float
    reporting_latency: float
    uploads: list[Upload]


@dataclass
class SEventDescription:
    """Class containing all the information required to create a S-event."""

    id: str
    source: str
    t_start: float
    t_0: float
    t_end: float
    gevent_ids: list[str]



"""A simple tool to display trees.

From https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python
"""
from __future__ import annotations

import re
from pathlib import Path
from typing import Any, Callable, Iterator

# prefix components:
SPACE = '    '
BRANCH = '│   '
# pointers:
TEE = '├── '
LAST = '└── '


def tree(
    dir_path: Path,
    criterion: Callable[[Path], bool] = lambda x: True,
    key: Callable[[Path], Any] | None = None,
    prefix: str = '',
) -> Iterator[tuple[Path, str]]:
    """A recursive generator to display a tree structure.

    Given a directory Path object it will yield a visual tree structure line by line
    with each line prefixed by the same characters.

    Parameters:
        dir_path: The directory to be displayed as a tree.
        criterion: Function that should return True to include the path in the tree.
        key: Function to sort the paths inside a given directory. By default, the
            alphanumerical order is used.
        prefix: Prefix to be prepended to the generated line.

    Yields:
        One line corresponding to a file or subdirectory contained in the specified
        path.
    """
    contents = sorted((_ for _ in dir_path.iterdir() if criterion(_)), key=key)
    # contents each get pointers that are ├── with a final └── :
    pointers = [TEE] * (len(contents) - 1) + [LAST]
    for pointer, path in zip(pointers, contents):
        yield path, prefix + pointer + path.name
        if path.is_dir():  # extend the prefix and recurse:
            extension = BRANCH if pointer == TEE else SPACE
            # i.e. space because last, └── , above so no more |
            yield from tree(path, criterion, key=key, prefix=prefix + extension)


SEVENT_REGEX = re.compile(r'M?S\d{5,}[a-z]+$')
GEVENT_REGEX = re.compile(r'[A-RT-Z]\d{5,}$')


def is_gevent(event_id: str) -> bool:
    """Returns true if the identifier is that of a G-event.

    Paramters:
        event_id: The event identifier.
    """
    return bool(GEVENT_REGEX.match(event_id))


def is_superevent(event_id: str) -> bool:
    """Returns true if the identifier is that of a superevent.

    A regular S-event identifier begins with `S`, a mock S-event with `MS`.

    Paramters:
        event_id: The event identifier.
    """
    return bool(SEVENT_REGEX.match(event_id))


def is_any_event(event_id: str) -> bool:
    """Returns true if the identifier is that of a G-event or S-event.

    Paramters:
        event_id: The event identifier.
    """
    return is_gevent(event_id) or is_superevent(event_id)


def split_filename(versioned_filename: str) -> tuple[str, int]:
    """Extract the file name and version from a versioned file.

    Parameters:
        versioned_filename: The input versioned file.

    Example:
        >>> split_filename('map.fits,3')
        ('map.fits', 3)
    """
    filename, version = versioned_filename.rsplit(',', 1)
    return filename, int(version)



import json
from pathlib import Path
from typing import Any

import pytest
from apischema import deserialize
from pytest_mock import MockerFixture

from mock_event_generator import GEventFetcher, SEventFetcher
from mock_event_generator.gracedbs import GraceDBAlias, GraceDBWithContext
from mock_event_generator.models import GEventDescription, SEventDescription

SEVENT_ID = 'S220609hl'
GEVENT_IDS = {'G587365', 'G587366', 'G587369'}
SOURCE = GraceDBWithContext.meg_from_alias_or_url(GraceDBAlias.MOCKED)


def test_fetchers(mocked_gracedb: None, tmp_path: Path) -> None:
    superevent_path = tmp_path / SEVENT_ID
    fetcher = SEventFetcher.from_id(SEVENT_ID, SOURCE, tmp_path)
    fetcher.fetch()
    description_path = superevent_path / 'description.json'
    assert description_path.is_file()
    description_serialized = json.loads(description_path.read_text())
    description_sevent = deserialize(SEventDescription, description_serialized)
    assert description_sevent.id == SEVENT_ID
    assert GEVENT_IDS.issubset(description_sevent.gevent_ids)

    for graceid in GEVENT_IDS:
        gevent_path = superevent_path / graceid
        assert gevent_path.is_dir()
        description_path = gevent_path / 'description.json'
        assert description_path.is_file()
        description_serialized = json.loads(description_path.read_text())
        description_gevent = deserialize(GEventDescription, description_serialized)
        assert description_gevent.id == graceid


@pytest.mark.parametrize('exception_cls', [KeyboardInterrupt, ValueError])
def test_cleanup(
    mocker: MockerFixture,
    mocked_gracedb: None,
    tmp_path: Path,
    exception_cls: type[BaseException],
) -> None:
    mocked_rmtree = mocker.patch('shutil.rmtree')
    mocked_fetch_no_cleanup = mocker.patch(
        'mock_event_generator.SEventFetcher._fetch_no_cleanup',
        side_effect=exception_cls,
    )

    with pytest.raises(exception_cls):
        SEventFetcher.from_id(SEVENT_ID, SOURCE, tmp_path).fetch()

    mocked_fetch_no_cleanup.assert_called_once()
    mocked_rmtree.assert_called_once()


def test_cleanup_present(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    mocked_rmtree = mocker.patch('shutil.rmtree')
    (tmp_path / SEVENT_ID).mkdir()
    SEventFetcher.from_id(SEVENT_ID, SOURCE, tmp_path).fetch()
    mocked_rmtree.assert_called_once()


def test_cleanup_missing(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    mocked_rmtree = mocker.patch('shutil.rmtree')
    SEventFetcher.from_id(SEVENT_ID, SOURCE, tmp_path).fetch()
    mocked_rmtree.assert_not_called()


def test_nologs(mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path) -> None:
    mocked_filter_logs = mocker.patch(
        'mock_event_generator.GEventFetcher._filter_logs', return_value=[]
    )
    SEventFetcher.from_id(SEVENT_ID, SOURCE, tmp_path).fetch()
    assert mocked_filter_logs.call_count == len(GEVENT_IDS)
    assert [p.name for p in (tmp_path / SEVENT_ID).glob('*')] == ['description.json']


def test_discard_offline(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    original_from_id = GEventFetcher.from_id

    def new_from_id(*args: Any, **keywords: Any) -> GEventFetcher:
        fetcher = original_from_id(*args, **keywords)
        fetcher.gevent['offline'] = True
        return fetcher

    mocker.patch(
        'mock_event_generator.GEventFetcher.from_id',
        autospec=True,
        side_effect=new_from_id,
    )

    SEventFetcher.from_id(SEVENT_ID, SOURCE, tmp_path).fetch()
    assert [p.name for p in (tmp_path / SEVENT_ID).glob('*')] == ['description.json']



import json
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from mock_event_generator.cache import (
    CacheEntry,
    EventFileCache,
    GEventCacheEntry,
    SEventCacheEntry,
)
from mock_event_generator.gracedbs import GraceDBAlias, GraceDBWithContext

GEVENT_ID = 'G587369'
SEVENT_ID = 'S220609hl'
SOURCE = GraceDBWithContext.meg_from_alias_or_url(GraceDBAlias.MOCKED)


def test_cache_gevent_enabled_missing(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    mocked_download = mocker.patch('mock_event_generator.cache.GEventFetcher.fetch')
    cache = EventFileCache(SOURCE, False, tmp_path)
    cache.get_gevent_cache_entry(GEVENT_ID)
    mocked_download.assert_called_once()


def test_cache_gevent_enabled_present(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    mocked_download = mocker.patch('mock_event_generator.cache.GEventFetcher.fetch')
    (tmp_path / GEVENT_ID).mkdir()
    cache = EventFileCache(SOURCE, False, tmp_path)
    cache.get_gevent_cache_entry(GEVENT_ID)
    mocked_download.assert_not_called()


def test_cache_gevent_disabled_missing(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    mocked_download = mocker.patch('mock_event_generator.cache.GEventFetcher.fetch')
    cache = EventFileCache(SOURCE, True, tmp_path)
    cache.get_gevent_cache_entry(GEVENT_ID)
    mocked_download.assert_called_once()


def test_cache_gevent_disabled_present(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    mocked_download = mocker.patch('mock_event_generator.cache.GEventFetcher.fetch')
    (tmp_path / GEVENT_ID).mkdir()
    cache = EventFileCache(SOURCE, True, tmp_path)
    cache.get_gevent_cache_entry(GEVENT_ID)
    mocked_download.assert_called_once()


def test_cache_sevent_enabled_missing(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    mocked_download = mocker.patch('mock_event_generator.cache.SEventFetcher.fetch')
    cache = EventFileCache(SOURCE, False, tmp_path)
    cache.get_sevent_cache_entry(SEVENT_ID)
    mocked_download.assert_called_once()


def test_cache_sevent_enabled_present(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    mocked_download = mocker.patch('mock_event_generator.cache.SEventFetcher.fetch')
    (tmp_path / SEVENT_ID).mkdir()
    cache = EventFileCache(SOURCE, False, tmp_path)
    cache.get_sevent_cache_entry(SEVENT_ID)
    mocked_download.assert_not_called()


def test_cache_sevent_disabled_missing(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    mocked_download = mocker.patch('mock_event_generator.cache.SEventFetcher.fetch')
    cache = EventFileCache(SOURCE, True, tmp_path)
    cache.get_sevent_cache_entry(SEVENT_ID)
    mocked_download.assert_called_once()


def test_cache_sevent_disabled_present(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    mocked_download = mocker.patch('mock_event_generator.cache.SEventFetcher.fetch')
    (tmp_path / SEVENT_ID).mkdir()
    cache = EventFileCache(SOURCE, True, tmp_path)
    cache.get_sevent_cache_entry(SEVENT_ID)
    mocked_download.assert_called_once()


def test_read_json(tmp_path: Path) -> None:
    filename = 'file.json'
    value = 'abc'
    (tmp_path / filename).write_text(json.dumps(value))
    cache_entry = CacheEntry(tmp_path)
    assert cache_entry.read_json(filename) == value


def test_gevent_cache_entry_from_sevent_cache_entry(tmp_path: Path) -> None:
    sevent_cache_entry = SEventCacheEntry(tmp_path)
    gevent_cache_entry = sevent_cache_entry / GEVENT_ID
    assert isinstance(gevent_cache_entry, GEventCacheEntry)
    assert gevent_cache_entry.path == tmp_path / GEVENT_ID


def test_gevent_cache_entry_from_sevent_cache_entry_error(tmp_path: Path) -> None:
    sevent_cache_entry = SEventCacheEntry(tmp_path)
    with pytest.raises(TypeError):
        sevent_cache_entry / 3  # type: ignore[operator]



"""List of the GraceDB instances."""
from __future__ import annotations

import re
from enum import Enum
from typing import Any

from ligo.gracedb.rest import GraceDb

from .exceptions import MEGInvalidGraceDBAliasOrURLError

GRACEDB_URLS = {
    'production': 'https://gracedb.ligo.org',
    'playground': 'https://gracedb-playground.ligo.org',
    'test': 'https://gracedb-test.ligo.org',
    'cnaf': 'https://gracedb-test01.igwn.org',
    'mocked': 'https://gracedb-donotexist.igwn.org',
}


class GraceDBWithContext(GraceDb):  # type: ignore[misc]
    """This class adds the URL and alias of the GranceDB server as attributes.

    The motivation of the meg_url attribute comes from the deprecation of GraceDb's
    service_url property.
    All the new attributes and methods are prefixed with 'meg_' to avoid potential
    future collision.
    """

    meg_url: str
    meg_alias: str | None
    MEG_URL_REGEX = re.compile(r'^https?://[A-Z0-9-.]+', re.I)

    def __init__(self, meg_url: str, meg_alias: str | None, **keywords: Any) -> None:
        """The GraceDBWithContext constructor."""
        super().__init__(**keywords)
        self.meg_url = meg_url
        self.meg_alias = meg_alias

    @classmethod
    def meg_from_alias_or_url(cls, value: str, **keywords: Any) -> GraceDBWithContext:
        """The GraceDBWithContext factory.

        Parameters:
            value: The GraceDB url, or an alias such as production, playground, etc.

        Raises:
            MEGInvalidGraceDBAliasOrURLError: When the specifier is not a known GraceDB
                alias nor a valid URL.
        """
        if value in GRACEDB_URLS:
            alias = value
            url = GRACEDB_URLS[value]
        elif cls.MEG_URL_REGEX.match(value):
            alias = None
            url = value
        else:
            raise MEGInvalidGraceDBAliasOrURLError(
                f'Invalid specifier for the GraceDB server: {value!r}. It should be a '
                f'URL or one of: {" ".join(GRACEDB_URLS)}',
            )
        return GraceDBWithContext(url, alias, service_url=f'{url}/api/', **keywords)


# XXX We should use StrEnum when mypy supports it: https://github.com/python/mypy/issues/11714  # noqa: E501
class GraceDBAlias(str, Enum):
    """Enum of the GraceDB instance names."""

    PRODUCTION = 'production'
    PLAYGROUND = 'playground'
    TEST = 'test'
    CNAF = 'cnaf'
    MOCKED = 'mocked'

    def __str__(self) -> str:
        """Value of the enum member."""
        return self.value



"""Exceptions specific to the mock_event_generator package."""

__all__ = [
    'MEGIgnoreEventError',
    'MEGInvalidGraceDBAliasOrURLError',
]


class MEGIgnoreEventError(Exception):
    """When the G-event should be filtered out."""


class MEGInvalidGraceDBAliasOrURLError(ValueError):
    """When the GraceDB specifier is not a known alias nor a valid URL."""



from __future__ import annotations

from pathlib import Path
from typing import Iterator

import pytest
from pytest_mock import MockerFixture
from trio.testing import MockClock

from mock_event_generator.cache import EventFileCache
from mock_event_generator.creators import (
    PIPELINE_GEVENT_CREATOR_CLASSES,
    CWBEventCreator,
    SEventCreator,
)
from mock_event_generator.gracedbs import GraceDBAlias, GraceDBWithContext

SEVENT_ID = 'S220609hl'
SOURCE = GraceDBWithContext.meg_from_alias_or_url(GraceDBAlias.MOCKED)
TARGET = GraceDBWithContext.meg_from_alias_or_url(GraceDBAlias.MOCKED)


@pytest.fixture
def cache(mocked_gracedb: None, tmp_path: Path) -> Iterator[EventFileCache]:
    yield EventFileCache(SOURCE, False, tmp_path)


@pytest.mark.parametrize('group', [None, 'Test'])
@pytest.mark.parametrize('search', [None, 'MDC'])
async def test_create_sevent(
    cache: EventFileCache,
    group: str | None,
    search: str | None,
    autojump_clock: MockClock,
) -> None:
    creator = SEventCreator.from_id(SEVENT_ID, SOURCE, cache)
    actual_gevents = await creator.create(group, search)
    sevent_cache_entry = cache.get_sevent_cache_entry(SEVENT_ID)
    gevent_ids = sevent_cache_entry.get_description().gevent_ids
    for gevent_id, actual_gevent in zip(gevent_ids, actual_gevents):
        description = (sevent_cache_entry / gevent_id).get_description()
        expected_group = group or description.group
        expected_search = search or description.search
        assert actual_gevent['group'] == expected_group
        assert actual_gevent['search'] == expected_search


async def test_create_sevent_unknown_pipeline(
    mocker: MockerFixture, caplog: pytest.LogCaptureFixture, cache: EventFileCache
) -> None:
    mocker.patch.dict(PIPELINE_GEVENT_CREATOR_CLASSES, clear=True)
    creator = SEventCreator.from_id(SEVENT_ID, SOURCE, cache)
    await creator.create(None, None)
    assert 'Cannot re-create G-event' in caplog.text


async def test_create_sevent_only_one_gevent(
    mocker: MockerFixture, cache: EventFileCache
) -> None:
    mocker.patch.dict(
        PIPELINE_GEVENT_CREATOR_CLASSES, {'CWB': CWBEventCreator}, clear=True
    )
    creator = SEventCreator.from_id(SEVENT_ID, SOURCE, cache)
    gevents = await creator.create(None, None, max_delay=1)
    assert len(gevents) == 1


async def test_create_sevent_single_upload() -> None:
    max_delay = 2.0
    assert SEventCreator._get_shrink_factor({'Z12345': [0]}, max_delay) == 1



from __future__ import annotations

from pathlib import Path

import pytest
from freezegun import freeze_time
from pytest_mock import MockerFixture
from typer.testing import CliRunner

from mock_event_generator.cache import EventFileCache
from mock_event_generator.cli import meg
from mock_event_generator.gracedbs import GraceDBAlias, GraceDBWithContext
from pytest_gracedb import PIPELINES

SEVENT_ID = 'S220609hl'
GEVENT_ID = PIPELINES['cwb'].gevent_id
SOURCE = GraceDBWithContext.meg_from_alias_or_url(GraceDBAlias.MOCKED)

runner = CliRunner()


def test_fetch_help(mocked_gracedb: None) -> None:
    result = runner.invoke(
        meg,
        [
            'fetch',
            '--help',
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert 'Fetch G-events and store them in the cache.' in result.output


@pytest.mark.parametrize('source', ['mocked', 'http://localhost', 'https://xyz'])
def test_fetch_gevent_alias(mocked_gracedb: None, source: str, tmp_path: Path) -> None:
    result = runner.invoke(
        meg,
        [
            'fetch',
            GEVENT_ID,
            '--source',
            source,
            '--cache-path',
            str(tmp_path),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0

    path = tmp_path / GEVENT_ID
    assert path.is_dir()
    actual_files = {_.name for _ in path.glob('*')} - {'description.json'}
    assert actual_files
    assert actual_files <= set(PIPELINES['cwb'].files)


def test_fetch_gevent_invalid_source(mocked_gracedb: None, tmp_path: Path) -> None:
    result = runner.invoke(
        meg,
        [
            'fetch',
            GEVENT_ID,
            '--source',
            'invalid',
            '--cache-path',
            str(tmp_path),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 1
    assert "Invalid specifier for the GraceDB server: 'invalid'" in result.output


def test_fetch_sevent(mocked_gracedb: None, tmp_path: Path) -> None:
    result = runner.invoke(
        meg,
        [
            'fetch',
            SEVENT_ID,
            '--source',
            'mocked',
            '--cache-path',
            str(tmp_path),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0

    path = tmp_path / SEVENT_ID
    assert path.is_dir()
    actual_files = {_.name for _ in path.glob('*')}
    assert 'description.json' in actual_files
    assert GEVENT_ID in actual_files


def test_fetch_invalid_event(mocked_gracedb: None, tmp_path: Path) -> None:
    result = runner.invoke(
        meg,
        [
            'fetch',
            'G12345',
            'XYZ1',
            'XYZ2',
            '--source',
            'mocked',
            '--cache-path',
            str(tmp_path),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 1
    assert "Invalid event identifier(s): 'XYZ1', 'XYZ2'." in result.output


def test_create_gevent(mocked_gracedb: None, tmp_path: Path) -> None:
    result = runner.invoke(
        meg,
        [
            'create',
            GEVENT_ID,
            '--target',
            'mocked',
            '--source',
            'mocked',
            '--cache-path',
            str(tmp_path),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert (tmp_path / GEVENT_ID).is_dir()


def test_create_gevent_invalid_target(mocked_gracedb: None, tmp_path: Path) -> None:
    result = runner.invoke(
        meg,
        [
            'create',
            GEVENT_ID,
            '--target',
            'invalid',
            '--source',
            'mocked',
            '--cache-path',
            str(tmp_path),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 1
    assert "Invalid specifier for the GraceDB server: 'invalid'" in result.output


def test_create_sevent(mocked_gracedb: None, tmp_path: Path) -> None:
    result = runner.invoke(
        meg,
        [
            'create',
            SEVENT_ID,
            '--target',
            'mocked',
            '--source',
            'mocked',
            '--cache-path',
            str(tmp_path),
            '--max-delay',
            '0',
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert (tmp_path / SEVENT_ID).is_dir()


def test_create_invalid_event(mocked_gracedb: None, tmp_path: Path) -> None:
    result = runner.invoke(
        meg,
        [
            'create',
            'G12345',
            'S12345a',
            'XYZ1',
            'XYZ2',
            '--target',
            'mocked',
            '--source',
            'mocked',
            '--cache-path',
            str(tmp_path),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 1
    assert "Invalid event identifier(s): 'XYZ1', 'XYZ2'." in result.output


@pytest.mark.parametrize(
    'include_files, expected_output',
    [
        (
            False,
            '├── G587366     MBTAOnline     CBC            AllSky         1338848303.869315\n    └── G587369     CWB            Burst          BBH            1338848303.7875',
        ),
        (
            True,
            '└── G587369     CWB            Burst          BBH            1338848303.7875\n        ├── description.json',
        ),
    ],
)
def test_cache_list(
    mocked_gracedb: None, tmp_path: Path, include_files: bool, expected_output: str
) -> None:
    cache = EventFileCache(SOURCE, False, tmp_path)
    cache.get_sevent_cache_entry(SEVENT_ID)
    assert len(list(tmp_path.glob('*'))) == 1

    result = runner.invoke(
        meg,
        [
            'cache',
            'list',
            '--include-files' if include_files else '--no-include-files',
            '--cache-path',
            str(tmp_path),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert expected_output in result.stdout


def test_cache_list_path_not_found(mocked_gracedb: None, tmp_path: Path) -> None:
    result = runner.invoke(
        meg,
        [
            'cache',
            'list',
            '--cache-path',
            str(tmp_path / 'invalid_path'),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 1
    assert 'Cache path does not exist' in result.output


@pytest.mark.parametrize('name', ['S00000abc', 'A00000', 'G00000', 'Z00000'])
def test_cache_clean_valid(mocked_gracedb: None, tmp_path: Path, name: str) -> None:
    (tmp_path / name).mkdir()
    result = runner.invoke(
        meg,
        [
            'cache',
            'clean',
            '--cache-path',
            str(tmp_path),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert len(list(tmp_path.glob('*'))) == 0


@pytest.mark.parametrize(
    'name, isdir',
    [
        ('G00000a', True),
        ('S00000', True),
        ('mystuff', True),
        ('otherstuff', False),
        ('G12345', False),
        ('S12345a', False),
    ],
)
def test_cache_clean_invalid(
    mocked_gracedb: None, tmp_path: Path, name: str, isdir: bool
) -> None:
    path = tmp_path / name
    if isdir:
        path.mkdir()
    else:
        path.touch()
    assert len(list(tmp_path.glob('*'))) == 1

    result = runner.invoke(
        meg,
        [
            'cache',
            'clean',
            '--cache-path',
            str(tmp_path),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert len(list(tmp_path.glob('*'))) == 1


def test_cache_clean_path_not_found(mocked_gracedb: None, tmp_path: Path) -> None:
    result = runner.invoke(
        meg,
        [
            'cache',
            'clean',
            '--cache-path',
            str(tmp_path / 'invalid_path'),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 1
    assert 'Cache path does not exist' in result.output


def test_ca_certificate(mocker: MockerFixture, tmp_path: Path) -> None:
    ca_bundle = tmp_path / 'cacert.pem'
    ca_bundle.touch()
    mocker.patch('certifi.where', return_value=str(ca_bundle))
    result = runner.invoke(
        meg,
        [
            'ca-certificate',
            'tests/data/certificate-terena-ssl-ca-3.pem',
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert ca_bundle.stat().st_size > 0


@freeze_time('3000/1/1')
def test_ca_certificate_expired(mocker: MockerFixture, tmp_path: Path) -> None:
    ca_bundle = tmp_path / 'cacert.pem'
    ca_bundle.touch()
    mocker.patch('certifi.where', return_value=str(ca_bundle))
    result = runner.invoke(
        meg,
        [
            'ca-certificate',
            'tests/data/certificate-terena-ssl-ca-3.pem',
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 1
    assert 'has expired' in result.output


def test_ca_certificate_already_added(mocker: MockerFixture, tmp_path: Path) -> None:
    ca_bundle = tmp_path / 'cacert.pem'
    ca_bundle.write_text(Path('tests/data/certificate-terena-ssl-ca-3.pem').read_text())
    mocker.patch('certifi.where', return_value=str(ca_bundle))
    result = runner.invoke(
        meg,
        [
            'ca-certificate',
            'tests/data/certificate-terena-ssl-ca-3.pem',
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert 'has already been added' in result.output



"""The mock_event_generator package.

This package downloads and caches existing G-Events from GraceDB and translates them
in time in order to produce new events, but with the same characteristics as the old
ones.
"""
import logging
from importlib.metadata import PackageNotFoundError, version

from .creators import GEventCreator, SEventCreator
from .fetchers import GEventFetcher, SEventFetcher

try:
    __version__ = version('mock_event_generator')
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    pass

__all__ = [
    'GEventCreator',
    'SEventCreator',
    'GEventFetcher',
    'SEventFetcher',
]

logging.basicConfig(
    format='%(asctime)s %(levelname)-8s %(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S',
)



from __future__ import annotations

from io import BytesIO
from pathlib import Path
from typing import Iterator

import lal
import pytest
from astropy.time import Time
from pytest_mock import MockerFixture
from requests import HTTPError

from ligo.gracedb.rest import GraceDb
from ligo.lw import lsctables
from ligo.lw import utils as ligolw_utils
from ligo.lw.ligolw import Document
from ligo.skymap.io.events.ligolw import ContentHandler
from mock_event_generator.cache import EventFileCache
from mock_event_generator.creators import (
    PIPELINE_GEVENT_CREATOR_CLASSES,
    CBCEventCreator,
    CWBEventCreator,
    ExternalEventCreator,
    GEventCreator,
    filter_uploads,
    filter_uploads_unless,
)
from mock_event_generator.gracedbs import GraceDBAlias, GraceDBWithContext
from mock_event_generator.models import Upload
from pytest_gracedb import PIPELINES

GEVENT_ID = 'G587369'
SOURCE = GraceDBWithContext.meg_from_alias_or_url(GraceDBAlias.MOCKED)
TARGET = GraceDBWithContext.meg_from_alias_or_url(GraceDBAlias.MOCKED)


@pytest.fixture
def cache(mocked_gracedb: None, tmp_path: Path) -> Iterator[EventFileCache]:
    yield EventFileCache(SOURCE, False, tmp_path)


@pytest.mark.parametrize('group', [None, 'Test'])
@pytest.mark.parametrize('search', [None, 'MDC'])
async def test_create_gevent(
    cache: EventFileCache, group: str | None, search: str | None
) -> None:
    creator = GEventCreator.from_id(GEVENT_ID, TARGET, cache)
    gevent = await creator.create(group, search, Time.now().gps, 0)
    expected_group = group or 'Burst'
    expected_search = search or 'BBH'
    assert gevent['group'] == expected_group
    assert gevent['search'] == expected_search


async def test_create_gevent_invalid_request(
    mocker: MockerFixture, cache: EventFileCache
) -> None:
    creator = GEventCreator.from_id(GEVENT_ID, TARGET, cache)
    mocker.patch(
        'pytest_gracedb.gracedb.TestDoubleGraceDB.create_event',
        return_value=(500, {}, '{"detail": "Reason 42"}'),
    )
    with pytest.raises(HTTPError) as excinfo:
        await creator.create(None, None, Time.now().gps, 0)
    exc: HTTPError = excinfo.value
    assert exc.response.status_code == 500
    assert 'Reason 42' in exc.response.text


def test_create_gevent_unknown_pipeline(
    mocker: MockerFixture, cache: EventFileCache
) -> None:
    mocker.patch.dict(PIPELINE_GEVENT_CREATOR_CLASSES, clear=True)
    with pytest.raises(NotImplementedError, match='Cannot re-create G-event'):
        GEventCreator.from_id(GEVENT_ID, TARGET, cache)


async def test_create_gevent_log_invalid_request(
    mocker: MockerFixture, cache: EventFileCache
) -> None:
    creator = GEventCreator.from_id(GEVENT_ID, TARGET, cache)
    mocker.patch(
        'pytest_gracedb.gracedb.TestDoubleGraceDB.create_event_log',
        return_value=(500, {}, '{"detail": "Reason 42"}'),
    )
    with pytest.raises(HTTPError) as excinfo:
        await creator.create(None, None, Time.now().gps, 0)
    exc: HTTPError = excinfo.value
    assert exc.response.status_code == 500
    assert 'Reason 42' in exc.response.text


async def test_create_gevent_label_invalid_request(
    mocker: MockerFixture, cache: EventFileCache
) -> None:
    creator = GEventCreator.from_id(GEVENT_ID, TARGET, cache)
    mocker.patch(
        'pytest_gracedb.gracedb.TestDoubleGraceDB.create_event_label',
        return_value=(500, {}, '{"detail": "Reason 42"}'),
    )
    with pytest.raises(HTTPError) as excinfo:
        await creator.create(None, None, Time.now().gps, 0)
    exc: HTTPError = excinfo.value
    assert exc.response.status_code == 500
    assert 'Reason 42' in exc.response.text


def test_time_shift_cwb() -> None:
    original_content = b"""noise:      6.227334e-24 9.412676e-24 1.561057e-23
segment:    1343218196.0000 1343218396.0000 1343218196.0000 1343218396.0000 1343218196.0000 1343218396.0000
start:      1343218303.7500 1343218303.7500 1343218303.7500
time:       1343218303.7875 1343218303.7939 1343218303.7985
stop:       1343218303.8750 1343218303.8750 1343218303.8750
inj_time:         0.0000       0.0000       0.0000"""
    delta_time = -1.5
    expected_content = b"""noise:      6.227334e-24 9.412676e-24 1.561057e-23
segment:    1343218196.0000 1343218396.0000 1343218196.0000 1343218396.0000 1343218196.0000 1343218396.0000
start:      1343218302.2500 1343218302.2500 1343218302.2500
time:       1343218302.2875 1343218302.2939 1343218302.2985
stop:       1343218302.3750 1343218302.3750 1343218302.3750
inj_time:         0.0000       0.0000       0.0000"""
    actual_content = CWBEventCreator._shift_original_data(original_content, delta_time)
    assert actual_content == expected_content


@pytest.mark.parametrize(
    'content',
    [
        b"""noise:      6.227334e-24 9.412676e-24 1.561057e-23
segment:    1343218196.0000 1343218396.0000 1343218196.0000 1343218396.0000 1343218196.0000 1343218396.0000
start:      1343218303.7500 1343218303.7500 1343218303.7500
stop:       1343218303.8750 1343218303.8750 1343218303.8750
inj_time:         0.0000       0.0000       0.0000""",
        b"""noise:      6.227334e-24 9.412676e-24 1.561057e-23
segment:    1343218196.0000 1343218396.0000 1343218196.0000 1343218396.0000 1343218196.0000 1343218396.0000
start:      1343218303.7500 1343218303.7500 1343218303.7500
time:       1343218303.7875
stop:       1343218303.8750 1343218303.8750 1343218303.8750
inj_time:         0.0000       0.0000       0.0000""",
        b"""noise:      6.227334e-24 9.412676e-24 1.561057e-23
segment:    1343218196.0000 1343218396.0000 1343218196.0000 1343218396.0000 1343218196.0000 1343218396.0000
start:      1343218303.7500 1343218303.7500 1343218303.7500
time:       1343218303.7875 1343218303.7939 1343218303.7985 1343218303.7985
stop:       1343218303.8750 1343218303.8750 1343218303.8750
inj_time:         0.0000       0.0000       0.0000""",
    ],
)
def test_time_shift_cwb_invalid_original_data1(content: bytes) -> None:
    with pytest.raises(ValueError, match='Could not extract the event times'):
        CWBEventCreator._shift_original_data(content, 0)


@pytest.mark.parametrize('pipeline_name', ['gstlal', 'mbta', 'pycbc', 'spiir'])
def test_time_shift_cbc(mocked_gracedb: None, pipeline_name: str) -> None:
    pipeline = PIPELINES[pipeline_name]
    data = GraceDb().files(pipeline.gevent_id, pipeline.files[0]).read()
    delta_time = 1.5
    doc: Document = ligolw_utils.load_fileobj(
        BytesIO(data), contenthandler=ContentHandler
    )
    new_doc: Document = ligolw_utils.load_fileobj(
        BytesIO(data), contenthandler=ContentHandler
    )
    CBCEventCreator._shift_original_document(new_doc, delta_time)

    table = lsctables.CoincInspiralTable.get_table(doc)
    new_table = lsctables.CoincInspiralTable.get_table(new_doc)
    for row, new_row in zip(table, new_table):
        assert new_row.end == row.end + delta_time

    table = lsctables.SnglInspiralTable.get_table(doc)
    new_table = lsctables.SnglInspiralTable.get_table(new_doc)
    for row, new_row in zip(table, new_table):
        assert new_row.end == row.end + delta_time
        assert new_row.end_time_gmst == lal.GreenwichMeanSiderealTime(
            row.end + delta_time
        )

    arrays = doc.getElementsByTagName('LIGO_LW')
    new_arrays = new_doc.getElementsByTagName('LIGO_LW')
    for array, new_array in zip(arrays, new_arrays):
        if not hasattr(array, 'Name') or array.Name != 'COMPLEX8TimeSeries':
            continue
        assert new_array.childNodes[0].pcdata == array.childNodes[0].pcdata + delta_time


def test_time_shift_external() -> None:
    content = b"""<?xml version='1.0' encoding='UTF-8'?>
<voe:VOEvent xmlns:voe="http://www.ivoa.net/xml/VOEvent/v2.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" ivorn="ivo://nasa.gsfc.gcn/Fermi#GBM_Alert_2019-08-17T21:21:20.05_584812348_1-240-TESTEVENT" role="observation" version="2.0" xsi:schemaLocation="http://www.ivoa.net/xml/VOEvent/v2.0  http://www.ivoa.net/xml/VOEvent/VOEvent-v2.0.xsd">
  <Who>
    <AuthorIVORN>ivo://nasa.gsfc.tan/gcn</AuthorIVORN>
    <Author>
      <shortName>Fermi (via VO-GCN)</shortName>
      <contactName>Julie McEnery</contactName>
      <contactPhone>+1-301-286-1632</contactPhone>
      <contactEmail>Julie.E.McEnery@nasa.gov</contactEmail>
    </Author>
    <Date>2022-03-17T14:37:25.840</Date>
    <Description>This VOEvent message was created with GCN VOE version: 1.42 11jun19</Description>
  </Who>
  <WhereWhen>
    <ObsDataLocation>
      <ObservatoryLocation id="GEOLUN"/>
      <ObservationLocation>
        <AstroCoordSystem id="UTC-FK5-GEO"/>
        <AstroCoords coord_system_id="UTC-FK5-GEO">
          <Time unit="s">
            <TimeInstant>
              <ISOTime>2022-03-17T14:37:25.840</ISOTime>
            </TimeInstant>
          </Time>
          <Position2D unit="deg">
            <Name1>RA</Name1>
            <Name2>Dec</Name2>
            <Value2>
              <C1>45.7900</C1>
              <C2>-28.0000</C2>
            </Value2>
            <Error2Radius>0.001</Error2Radius>
          </Position2D>
        </AstroCoords>
      </ObservationLocation>
    </ObsDataLocation>
  <Description>The RA,Dec coordinates are of the type: unavailable/inappropriate.</Description>
  </WhereWhen>
</voe:VOEvent>
"""
    new_content = ExternalEventCreator._shift_original_data(content, 15.0)
    assert b'<Date>2022-03-17T14:37:40.840</Date>' in new_content
    assert b'<ISOTime>2022-03-17T14:37:40.840</ISOTime>' in new_content


def test_time_shift_external_missing() -> None:
    content = b"""<?xml version='1.0' encoding='UTF-8'?>
<voe:VOEvent xmlns:voe="http://www.ivoa.net/xml/VOEvent/v2.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" ivorn="ivo://nasa.gsfc.gcn/Fermi#GBM_Alert_2019-08-17T21:21:20.05_584812348_1-240-TESTEVENT" role="observation" version="2.0" xsi:schemaLocation="http://www.ivoa.net/xml/VOEvent/v2.0  http://www.ivoa.net/xml/VOEvent/VOEvent-v2.0.xsd">
  <Who>
    <AuthorIVORN>ivo://nasa.gsfc.tan/gcn</AuthorIVORN>
    <Author>
      <shortName>Fermi (via VO-GCN)</shortName>
      <contactName>Julie McEnery</contactName>
      <contactPhone>+1-301-286-1632</contactPhone>
      <contactEmail>Julie.E.McEnery@nasa.gov</contactEmail>
    </Author>
    <Date>2022-03-17T14:37:25.840</Date>
    <Description>This VOEvent message was created with GCN VOE version: 1.42 11jun19</Description>
  </Who>
</voe:VOEvent>
"""
    with pytest.raises(
        ValueError,
        match=r"No element specified by '\./WhereWhen/ObsDataLocation/ObservationLocation/AstroCoords/Time/TimeInstant/ISOTime'",
    ):
        ExternalEventCreator._shift_original_data(content, 0.0)


@pytest.mark.parametrize(
    'upload',
    [
        pytest.param(
            Upload('', set(), 'em_bright.json,0', 0.0), marks=pytest.mark.xfail
        ),
        Upload('', set(), 'p_astro.json,0', 0.0),
        pytest.param(
            Upload('', {'em_bright'}, 'em_bright.json,0', 0.0), marks=pytest.mark.xfail
        ),
        Upload('', {'p_astro'}, 'p_astro.json,0', 0.0),
        Upload('', {'sky_loc'}, 'localization.fits,0', 0.0),
        Upload('', {'sky_loc'}, 'localization.fits.gz,0', 0.0),
    ],
)
async def test_write_label_in(
    mocker: MockerFixture, cache: EventFileCache, upload: Upload
) -> None:
    creator = GEventCreator.from_id(GEVENT_ID, TARGET, cache)
    mocked_write_label = mocker.patch.object(creator.target, 'write_label')
    await creator._write_label(creator.gevent.id, upload)
    mocked_write_label.assert_called_once()


@pytest.mark.parametrize(
    'upload',
    [
        Upload('', set(), 'em_bright.json,1', 0.0),
        Upload('', set(), 'p_astro.json,1', 0.0),
        Upload('', {'em_bright'}, 'em_bright.json,1', 0.0),
        Upload('', {'p_astro'}, 'p_astro.json,1', 0.0),
        Upload('', set(), 'localization.fits,0', 0.0),
        Upload('', set(), 'localization.fits.gz,0', 0.0),
        Upload('', {'A'}, 'localization.fits,0', 0.0),
        Upload('', {'A'}, 'localization.fits.gz,0', 0.0),
        Upload('', {'sky_loc'}, 'localization.fits,1', 0.0),
        Upload('', {'sky_loc'}, 'localization.fits.gz,1', 0.0),
        Upload('', {'sky_loc'}, 'localization.fit,0', 0.0),
        Upload('', {'sky_loc'}, 'localization.fit.gz,0', 0.0),
    ],
)
async def test_write_label_out(
    mocker: MockerFixture, cache: EventFileCache, upload: Upload
) -> None:
    creator = GEventCreator.from_id(GEVENT_ID, TARGET, cache)
    mocked_write_label = mocker.patch.object(creator.target, 'write_label')
    await creator._write_label(creator.gevent.id, upload)
    mocked_write_label.assert_not_called()


def test_filter_uploads_include_all_files() -> None:
    uploads = [Upload('', set(), 'file.txt', 0.0)]
    assert filter_uploads_unless(uploads, True) == uploads


@pytest.mark.parametrize(
    'upload',
    [
        Upload('', {'em_bright'}, 'em_bright.json,0', 0.0),
        Upload('', {'p_astro'}, 'p_astro.json,0', 0.0),
        Upload('', {'sky_loc'}, 'localization.fits,0', 0.0),
        Upload('', {'sky_loc'}, 'localization.fits.gz,0', 0.0),
        Upload('', {'em_bright'}, 'em_bright.json,1', 0.0),
        Upload('', {'p_astro'}, 'p_astro.json,1', 0.0),
        Upload('', {'sky_loc'}, 'localization.fits,1', 0.0),
        Upload('', {'sky_loc'}, 'localization.fits.gz,1', 0.0),
        Upload('', {'em_bright'}, 'file.json,0', 0.0),
        Upload('', {'p_astro'}, 'file.json,0', 0.0),
        Upload('', {'sky_loc'}, 'file.txt,0', 0.0),
        Upload('', {'sky_loc'}, 'file.txt.gz,0', 0.0),
    ],
)
def test_filter_uploads_in(upload: Upload) -> None:
    uploads = [Upload('', set(), 'original_data.txt,0', 0.0), upload]
    assert filter_uploads(uploads) == uploads


@pytest.mark.parametrize(
    'upload',
    [
        Upload('', {'A'}, 'file.txt,0', 0.0),
        Upload('', {'em_bright'}, 'em_bright.png,0', 0.0),
        Upload('', {'p_astro'}, 'p_astro.png,0', 0.0),
        Upload('', {'sky_loc'}, 'localization.png,0', 0.0),
    ],
)
def test_filter_uploads_out(upload: Upload) -> None:
    uploads = [Upload('', set(), 'original_data.txt,0', 0.0), upload]
    assert filter_uploads(uploads) == [uploads[0]]



"""Classes for downloading and caching GraceDB G-events and S-Events."""
from __future__ import annotations

import json
import logging
import os
import re
import shutil
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from tempfile import mkdtemp
from typing import Any, cast

import urllib3
from apischema import serialize
from requests import HTTPError

from .exceptions import MEGIgnoreEventError
from .gracedbs import GraceDBAlias, GraceDBWithContext
from .models import GEventDescription, SEventDescription, Upload

__all__ = [
    'GEventFetcher',
    'SEventFetcher',
]

logger = logging.getLogger(__name__)

DEFAULT_SOURCE_GRACEDB = GraceDBAlias.PRODUCTION
EMFOLLOW_REGEX = re.compile(
    r'(Log File Created|Coinc Table Created|.* created by GraceDbPlotter)$'
)


@dataclass
class GEventFetcher:
    """Class to fetch and store files associated to a G-event.

    Attributes:
        gevent: The G-event, as returned by GraceDB.
        path: The directory where the event files will be stored.
        gracedb: The GraceDB client instance used for the queries.
    """

    gevent: dict[str, Any]
    source: GraceDBWithContext
    path: Path

    @property
    def id(self) -> str:
        """Returns the G-event GraceDB identifier."""
        return cast(str, self.gevent['graceid'])

    @classmethod
    def from_id(
        cls,
        gevent_id: str,
        source: GraceDBWithContext,
        path: Path,
    ) -> GEventFetcher:
        """The GEventFetcher factory.

        Return the GEventFetcher associated to an event id.

        Parameters:
            gevent_id: The G-event GraceDB identifier.
            source: The GraceDB client from which events are downloaded.
            path: The directory where the event files will be stored.
        """
        gevent = source.event(gevent_id).json()
        return GEventFetcher(gevent, source, path)

    def fetch(self) -> None:
        """Download all the information required to re-create the event.

        The files are first written in a temporary directory, and this directory
        is renamed into the G-event identifier only upon success, in order to avoid
        dealing with incomplete downloads.
        """
        self.path.mkdir(exist_ok=True)
        tmp_path = Path(mkdtemp(dir=self.path))
        try:
            self._fetch_no_cleanup(tmp_path)

            target = self.path / self.id
            if target.exists():
                shutil.rmtree(target)
            os.rename(tmp_path, target)
        except BaseException:
            shutil.rmtree(tmp_path)
            raise

    def _fetch_no_cleanup(self, tmp_path: Path) -> None:
        description = self._get_description()
        self._store_description(description, tmp_path)
        filenames = [_.filename for _ in description.uploads]
        self._store_files(tmp_path, filenames)

    def _get_description(self) -> GEventDescription:
        description = GEventDescription(
            id=self.id,
            source=self.source.meg_alias or self.source.meg_url,
            pipeline=self.gevent['pipeline'],
            group=self.gevent['group'],
            search=self.gevent['search'],
            offline=self.gevent['offline'],
            gpstime=self.gevent['gpstime'],
            reporting_latency=self.gevent['reporting_latency'],
            uploads=self._get_uploads(),
        )
        return description

    def _store_description(self, description: GEventDescription, path: Path) -> None:
        """Cache the G-event description as a json file.

        This file contains all the information to re-create an event.
        """
        description_path = path / 'description.json'
        description_path.write_text(json.dumps(serialize(description), indent=4))

    def _get_uploads(self) -> list[Upload]:
        full_logs = self._get_logs()
        logs = self._filter_logs(full_logs)
        if not logs:
            raise MEGIgnoreEventError(
                f'No pipeline logs for the event {self.id!r}:\n'
                f'{chr(13).join(str(_) for _ in full_logs)}'
            )
        t0 = self._convert_created(logs[0]['created'])
        return [self._get_upload(log, t0) for log in logs]

    def _get_upload(self, log: dict[str, Any], t0: datetime) -> Upload:
        delay = (self._convert_created(log['created']) - t0).total_seconds()
        upload = Upload(
            message=log['comment'],
            tags=log['tag_names'],
            filename=f'{log["filename"]},{log["file_version"]}',
            delay=delay,
        )
        return upload

    def _get_logs(self) -> list[dict[str, Any]]:
        """Return the event logs."""
        response = self.source.logs(self.id)
        response.raise_for_status()
        response_logs = response.json()
        if response_logs['links']['first'] != response_logs['links']['last']:
            raise NotImplementedError('Logs are paginated.')
        logs: list[dict[str, Any]] = response_logs['log']
        return logs

    def _filter_logs(self, logs: list[dict[str, Any]]) -> list[dict[str, Any]]:
        """Filter out logs not issued by the pipeline.

        The discarded logs are those:
        - with no file
        - not issued by the pipeline
        - issued by 'emfollow' except for the first upload
        - with a comment such as 'Log File Created', 'Coinc Table Created' or
            '... created by GraceDbPlotter'.

        Parameters:
            logs: The list of logs associated to the current event as returned by
                GraceDB.
        """
        filtered_logs = [
            _
            for _ in logs
            if _['file']
            and _['issuer'] == self.gevent['submitter']
            and not (_['issuer'] == 'emfollow' and _['N'] > 1)
            and not EMFOLLOW_REGEX.match(_['comment'])
        ]
        return filtered_logs

    @staticmethod
    def _convert_created(dt: str) -> datetime:
        """Convert the log's field 'created' into a datetime.

        Parameters:
            dt: The time string, as returned by GraceDB.

        Example:
            >>> GEventFetcher._convert_created('2020-02-25 06:04:45 UTC')
            datetime.datetime(2020, 2, 25, 5, 4, 45, tzinfo=datetime.timezone.utc)
        """
        return datetime.strptime(dt, '%Y-%m-%d %H:%M:%S UTC').astimezone(timezone.utc)

    def _store_files(self, path: Path, filenames: list[str]) -> None:
        """Fetch and cache the event's files required to re-generate it.

        Parameters:
            path: The path of the directory where the files will be stored.
            filenames: The file names to be downloaded and stored.
        """
        for filename in filenames:
            self._store_file(path, filename)

    def _store_file(self, path: Path, filename: str) -> None:
        """Fetch and cache an event file.

        Parameters:
            path: The path of the directory where the file will be stored.
            filename: The file name to be downloaded and stored.

        Raises:
            ValueError: When the file does not exist for the current event.
        """
        try:
            response = self.source.files(self.id, filename)
        except HTTPError as exc:
            if exc.response.status != 404:
                raise
            if filename.endswith('.gz'):
                self._store_file(path, filename[:-3])
                return
            raise ValueError(f'G-event {self.id}: File {filename!r} not found.')

        assert isinstance(response, urllib3.response.HTTPResponse)
        (path / filename).write_bytes(response.read())


@dataclass
class SEventFetcher:
    """Class to fetch and store files associated to a super event.

    Attributes:
        sevent: The S-event, as returned by GraceDB.
        path: The directory where the event files will be stored.
        gracedb: The GraceDB client instance used for the queries.
    """

    sevent: dict[str, Any]
    source: GraceDBWithContext
    path: Path

    @property
    def id(self) -> str:
        """Returns the S-event GraceDB identifier."""
        return cast(str, self.sevent['superevent_id'])

    @classmethod
    def from_id(
        cls,
        sevent_id: str,
        source: GraceDBWithContext,
        path: Path,
    ) -> SEventFetcher:
        """The SEventFetcher factory.

        Return the SEventFetcher associated to a super event id.

        Parameters:
            sevent_id: The S-event GraceDB identifier.
            source: The GraceDB client from which events are downloaded.
            path: The directory where the event files will be stored.
        """
        sevent = source.superevent(sevent_id).json()
        return SEventFetcher(sevent, source, path)

    def fetch(self) -> None:
        """Download all the information required to re-create the super-event.

        The files are first written in a temporary directory, and this directory
        is renamed into the G-event identifier only upon success, in order to avoid
        dealing with incomplete downloads.
        """
        self.path.mkdir(exist_ok=True)
        tmp_path = Path(mkdtemp(dir=self.path))
        try:
            self._fetch_no_cleanup(tmp_path)

            target = self.path / self.id
            if target.exists():
                shutil.rmtree(target)
            os.rename(tmp_path, target)
        except BaseException:
            shutil.rmtree(tmp_path)
            raise

    def _fetch_no_cleanup(self, tmp_path: Path) -> None:
        description = self._get_description()
        valid_ids = []
        for gevent_id in description.gevent_ids:
            fetcher = GEventFetcher.from_id(gevent_id, self.source, tmp_path)
            if fetcher.gevent['offline']:
                continue
            logger.info(f'Downloading {gevent_id}...')
            try:
                fetcher.fetch()
            except MEGIgnoreEventError as exc:
                logger.warning(f'Aborting {gevent_id}: {exc}')
                continue
            valid_ids.append(gevent_id)

        description.gevent_ids = valid_ids
        self._store_description(description, tmp_path)

    def _get_description(self) -> SEventDescription:
        description = SEventDescription(
            id=self.id,
            source=self.source.meg_alias or self.source.meg_url,
            t_start=self.sevent['t_start'],
            t_0=self.sevent['t_0'],
            t_end=self.sevent['t_end'],
            gevent_ids=self.sevent['gw_events'],
        )
        return description

    def _store_description(self, description: SEventDescription, path: Path) -> None:
        """Cache the description as a JSON file."""
        description_path = path / 'description.json'
        description_path.write_text(json.dumps(serialize(description), indent=4))



from base64 import b64decode

import responses
from requests import Response

from mock_event_generator.gracedbs import GRACEDB_URLS, GraceDBWithContext


def test_gracedb_basic_authentication() -> None:
    call_count = 0

    def callback(resp: Response) -> Response:
        nonlocal call_count
        call_count += 1
        headers = resp.request.headers
        assert 'Authorization' in headers
        decoded = b64decode(headers['Authorization'].removeprefix('Basic '))
        assert decoded.decode() == f'{username}:{password}'
        return resp

    username = 'username'
    password = 'password'
    alias = 'mocked'
    url = GRACEDB_URLS[alias]
    client = GraceDBWithContext(alias, url, username=username, password=password)
    api = GRACEDB_URLS[alias] + '/api/'
    with responses.RequestsMock(response_callback=callback) as mock:
        mock.add(responses.GET, api)
        client.get(api)
    assert call_count == 1
